from datetime import datetime, timedelta


class TimeLog:
    """
    Simple class to keep track of total time spend doing
    certain events. Can add events when TimeLog is inherited.
    """

    def __init__(self, timed_events):
        self.times = {}
        for e in timed_events:
            self.times[e] = timedelta(0)

    def __repr__(self):
        maxlen = max([len(l) for l in list(self.times)])
        lines = ["Total Time in Seconds:"]
        for key, time in self.times.items():
            space_add = maxlen - len(key)
            lines.append(f"""{key}: {' '*space_add}{"%.2f" % time.total_seconds()}""")

        return "\n".join(lines)

    def add_time(self, event, s):
        """
        Add the time since s to the total time of event.
        """
        self.times[event] += datetime.now() - s

    def add_events(self, timed_events):
        """
        Add events. Used for inheritance
        """
        for e in timed_events:
            self.times[e] = timedelta(0)

    def get_event_time(self, event):
        self.times[event].total_seconds()



from .ImputedDataSet import ImputedDataSet
from .MultipleImputedDataSet import MultipleImputedDataSet
from .KernelDataSet import KernelDataSet
from pandas import DataFrame
from typing import Union, List, Dict
from .utils import ensure_rng


class MultipleImputedKernel(MultipleImputedDataSet):
    """
    Multiple Imputed Kernel

    Creates and stores a collection of KernelDataSet instances.
    Has methods that allow for easy access to datasets, as well
    as comparison and plotting methods.

    More details on usage can be found on the GitHub:
    https://github.com/AnotherSamWilson/miceforest

    Parameters
    ----------
    data: DataFrame
        A pandas DataFrame to impute.
    datasets: int, optional(default=5)
        The number of kernel datasets to create.
    variable_schema: None or list or Dict[str, List[str]]
        If None all variables are used to impute all variables which have
        missing values.
        If list all variables are used to impute the variables in the list
        If dict the values will be used to impute the keys.
    save_all_iterations: boolean, optional(default=False)
        Save all iterations that have been imputed, or just the latest.
        Saving all iterations allows for additional plotting,
        but may take more memory

    Methods
    -------
    append()
        Adds an imputed dataset.

    remove()
        Removes an imputed dataset.

    dataset_count()
        Returns the number of datasets being stored.

    get_correlations()
        Returns the pearson correlations between imputed datasets.

    plot_correlations()
        Plots the pearson correlations between imputed datasets.

    plot_imputed_distributions()
        Plots the distribution of imputed values against the original.

    iteration_count()
        Returns the iterations that have been run to impute the data.

    complete_data()
        Returns the completed data for the specified dataset.

    plot_mean_convergence()
        Plots the means of the imputed values for each dataset over
        iterations.

    impute_new_data()
        Imputes a new dataset using the models obtained from performing
        mice on the kernel dataset. Pulls mean matching candidate values
        from the kernel data.

    mice()
        Performs mice on the kernel data.

    get_feature_importance()
        Returns feature importance of the specified dataset.

    plot_feature_importance()
        Plots the feature importance of the specified dataset.

    """

    def __init__(
        self,
        data: DataFrame,
        datasets: int = 5,
        variable_schema: Union[List[str], Dict[str, List[str]]] = None,
        mean_match_candidates: Union[int, Dict[str, int]] = None,
        save_all_iterations: bool = False,
        save_models: int = 1,
        random_state=None,
    ):

        random_state = ensure_rng(random_state)

        super().__init__(
            initial_dataset=KernelDataSet(
                data=data,
                variable_schema=variable_schema,
                mean_match_candidates=mean_match_candidates,
                save_all_iterations=save_all_iterations,
                save_models=save_models,
                random_state=random_state,
            )
        )

        # Prime with the required number of kernel datasets
        while self.dataset_count() < datasets:
            self.append(
                KernelDataSet(
                    data=data,
                    variable_schema=variable_schema,
                    mean_match_candidates=self.mean_match_candidates,
                    save_all_iterations=save_all_iterations,
                    save_models=save_models,
                    random_state=random_state,
                )
            )

        self.save_models = save_models
        self.iteration_time_seconds = 0

    def __repr__(self):
        mss = {0: "None", 1: "Last Iteration", 2: "All Iterations"}
        summary_string = (
            f"""\
              Class: MultipleImputedKernel
       Models Saved: {mss[self.save_models]}\n"""
            + self._mids_info()
        )
        return summary_string

    def mice(self, iterations: int = 5, verbose: bool = False, **kw_fit):
        """
        Calls mice() on all datasets stored in this instance.

        Multiple Imputation by Chained Equations (MICE) is an
        iterative method which fills in (imputes) missing data
        points in a dataset by modeling each column using the
        other columns, and then inferring the missing data.

        For more information on MICE, and missing data in
        general, see Stef van Buuren's excellent online book:
        https://stefvanbuuren.name/fimd/ch-introduction.html

        For detailed usage information, see this project's
        README on the github repository:
        https://github.com/AnotherSamWilson/miceforest#The-MICE-Algorithm



        Parameters
        ----------
        iterations: int
            The number of iterations to run.
        verbose: bool
            Should information about the process
            be printed?
        kw_fit:
            Additional arguments to pass to
            sklearn.RandomForestRegressor and
            sklearn.RandomForestClassifier
        """
        for dataset in list(self.keys()):
            if verbose:
                print("Dataset " + str(dataset))
            self[dataset].mice(iterations=iterations, verbose=verbose, **kw_fit)

    def impute_new_data(
        self,
        new_data: DataFrame,
        datasets: List[int] = None,
        iterations: int = None,
        save_all_iterations: bool = False,
        verbose: bool = False,
    ) -> Union[ImputedDataSet, MultipleImputedDataSet]:
        """
        Call impute_new_data on multiple kernel kernel datasets,
        returning a MultipleImputedDataset.
        If len(datasets) == 1, an ImputedDataSet is returned.

        Parameters
        ----------
        new_data: pandas DataFrame
            The new data to impute
        datasets: None, List[int]
            The datasets and corresponding models from
            the kernel to use for imputation. If None,
            all datasets are used.
        iterations: None, int
            The iterations to run. If None, the same number
            of iterations in the kernel are used.
        save_all_iterations: bool
            Whether to save about all of the imputed values
            in every iteration, or just the last iteration.
        verbose: bool
            Print progress

        Returns ImputedDataSet or MultipleImputedDataSet
        -------
            If 1 dataset is selected, an ImputedDataSet is returned
            If more than 1 dataset is selected, a MultipleImputedDataset
            is returned.

        """
        if datasets is None:
            datasets = list(self.keys())

        # Create an ImputedDataSet on the first dataset in the list
        if verbose:
            print("Dataset " + str(datasets[0]))
        imputed_data_set = self[datasets.pop(0)].impute_new_data(
            new_data=new_data,
            iterations=iterations,
            save_all_iterations=save_all_iterations,
            verbose=verbose,
        )
        if len(datasets) > 0:
            multiple_imputed_set = MultipleImputedDataSet(
                initial_dataset=imputed_data_set
            )
            while len(datasets) > 0:
                if verbose:
                    print("Dataset " + str(datasets[0]))
                multiple_imputed_set.append(
                    self[datasets.pop(0)].impute_new_data(
                        new_data=new_data,
                        iterations=iterations,
                        save_all_iterations=save_all_iterations,
                        verbose=verbose,
                    )
                )
            return multiple_imputed_set
        else:
            return imputed_data_set

    def get_feature_importance(self, dataset: int = 0) -> DataFrame:
        """
        Return a dataframe of feature importance values.
        The values represent the scaled importance of the
        column variables in imputing the row variables.

        Parameters
        ----------
        dataset: int
            The dataset to get the feature importance of.

        Returns
        -------

        """
        return self[dataset].get_feature_importance()

    def plot_feature_importance(self, dataset: int = 0, **kw_plot):
        """
        Plot the feature importance of a specific stored dataset.

        Parameters
        ----------
        dataset: int
            The dataset to plot the feature importance of.
        kw_plot
            Other arguments passed to seaborn.heatmap()

        Returns
        -------

        """
        self[dataset].plot_feature_importance(**kw_plot)



from .ImputedDataSet import ImputedDataSet
from sklearn.neighbors import NearestNeighbors
from .TimeLog import TimeLog
from datetime import datetime
from .utils import _get_default_mmc, _default_rf_classifier, _default_rf_regressor
from pandas import DataFrame
import numpy as np
from typing import Union, List, Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

_TIMED_EVENTS = ["mice", "model_fit", "model_predict", "mean_match", "impute_new_data"]


class KernelDataSet(ImputedDataSet):
    """
    Creates a kernel dataset. This dataset can:
        - Perform MICE on itself
        - Impute new data from models obtained from MICE.

    Parameters
    ----------
    data: DataFrame
        A pandas DataFrame to impute.

    variable_schema: None or list or dict
        If None all variables are used to impute all variables which have
        missing values.
        If list all variables are used to impute the variables in the list
        If dict the values will be used to impute the keys.

    mean_match_candidates:  None or int or dict
        The number of mean matching candidates to use.
        Candidates are _always_ drawn from a kernel dataset, even
        when imputing new data.

        Mean matching follows the following rules based on variable type:
            Categorical:
                If mmc = 0, the predicted class is used
                If mmc > 0, return class based on random draw weighted by
                    class probability for each sample.
            Numeric:
                If mmc = 0, the predicted value is used
                If mmc > 0, obtain the mmc closest candidate
                    predictions and collect the associated
                    real candidate values. Choose 1 randomly.

        For more information, see:
        https://github.com/AnotherSamWilson/miceforest#Predictive-Mean-Matching

    save_all_iterations: boolean, optional(default=True)
        Save all the imputation values from all iterations, or just
        the latest. Saving all iterations allows for additional
        plotting, but may take more memory

    save_models: int
        Which models should be saved:
            = 0: no models are saved. Cannot get feature importance or
                impute new data.
            = 1: only the last model iteration is saved. Can only get
                feature importance of last iteration. New data is
                imputed using the last model for all specified iterations.
                This is only an issue if data is heavily Missing At Random.
            = 2: all model iterations are saved. Can get feature importance
                for any iteration. When imputing new data, each iteration is
                imputed using the model obtained at that iteration in mice.
                This allows for imputations that most closely resemble those
                that would have been obtained in mice.

    random_state: None,int, or numpy.random.RandomState
        Ensures a random state throughout the process

    Methods
    -------
    iteration_count()
        Returns the number of iterations that have been run to impute
        the data.

    complete_data()
        Returns a completed dataset, with missing values imputed.

    get_means()
        Returns the average value of numeric variables at each iteration.

    plot_mean_convergence()
        Plots the average value of numeric variables at each iteration.

    get_model()
        Returns the model used to impute the specified variable.

    mice()
        Performs mice on the missing data.

    impute_new_data()
        Uses the model obtained from mice to impute a new dataset.

    get_feature_importance()
        Returns feature importance of the specified dataset.

    plot_feature_importance()
        Plots the feature importance of the specified dataset.

    """

    def __init__(
        self,
        data: DataFrame,
        variable_schema: Union[List[str], Dict[str, List[str]]] = None,
        mean_match_candidates: Union[int, Dict[str, int]] = None,
        save_all_iterations: bool = True,
        save_models: int = 1,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        super().__init__(
            data=data,
            variable_schema=variable_schema,
            mean_match_candidates=mean_match_candidates,
            save_all_iterations=save_all_iterations,
            random_state=random_state,
        )

        self.save_models = save_models

        # Format mean_match_candidates before priming datasets
        available_candidates = {
            var: self.data_shape[0] - self.data[var].isna().sum()
            for var in self.response_vars
        }

        if self.mean_match_candidates is None:
            self.mean_match_candidates = {
                var: _get_default_mmc(available_candidates[var])
                for var in self.response_vars
            }

        elif isinstance(self.mean_match_candidates, int):
            self.mean_match_candidates = {
                key: self.mean_match_candidates for key in self.response_vars
            }
        elif isinstance(mean_match_candidates, dict):
            if not set(mean_match_candidates) == set(self.response_vars):
                raise ValueError(
                    "mean_match_candidates not consistent with variable_schema. "
                    + "Do all variables in variable_schema have missing values?."
                )

        mmc_inadequate = [
            var
            for var, mmc in self.mean_match_candidates.items()
            if (mmc >= available_candidates[var])
        ]
        if len(mmc_inadequate) > 0:
            raise ValueError(
                "<"
                + ",".join(mmc_inadequate)
                + ">"
                + " do not have enough candidates to perform mean matching."
            )

        # Initialize models
        self.models: Dict[str, Dict] = {var: {0: None} for var in self.response_vars}

        self.time_log = TimeLog(_TIMED_EVENTS)

    def __repr__(self):
        summary_string = " " * 14 + "Class: KernelDataSet\n" + self._ids_info()
        return summary_string

    def _insert_new_model(self, var, model):
        """
        Inserts a new model if save_mdoels > 0.
        Deletes the prior one if save_models == 1.
        """
        current_iter = self.iteration_count(var)
        if self.save_models == 0:
            return
        else:
            self.models[var][current_iter + 1] = model
        if self.save_models == 1:
            del self.models[var][current_iter]

    def get_model(self, var, iteration=None):
        """
        Return the model for a specific variable, iteration.

        Parameters
        ----------
        var: str
            The variable that was imputed
        iteration: int
            The model iteration to return. Keep in mind if:
                - save_models == 0, no models are saved
                - save_models == 1, only the last model iteration is saved
                - save_models == 2, all model iterations are saved

        Returns: RandomForestRegressor or RandomForestClassifier
            The model used to impute this specific variable, iteration.

        """

        iteration = self._default_iteration(iteration)
        try:
            return self.models[var][iteration]
        except Exception:
            raise ValueError("Iteration was not saved")

    def _mean_match(
        self,
        model,
        mmc: int,
        is_categorical: bool,
        bachelor_features: DataFrame,
        candidate_features: DataFrame = None,
        candidate_values: np.array = None,
    ):
        """
        Performs mean matching. Logic:
            if categorical:
                Return class based on random draw weighted by
                class probability for each sample.
            if numeric:
                For each sample prediction, obtain the mmc closest
                candidate_values and collect the associated
                candidate_features. Choose 1 randomly.

        For a graphical example, see:
        https://github.com/AnotherSamWilson/miceforest#Predictive-Mean-Matching

        Parameters
        ----------
        model
            The model
        mmc
            The mean matching candidates
        is_categorical
            Is the feature we are imputing categorical
        bachelor_features
            The features of the variable to be imputed
        candidate_features
            The features of the candidates
        candidate_values
            The real values of the candidates

        Returns
        -------
            An array of imputed values.

        """
        mean_match_s = datetime.now()

        if is_categorical:
            # Select category according to their probability for each sample
            model_predict_s = datetime.now()
            bachelor_preds = model.predict_proba(bachelor_features)
            self.time_log.add_time("model_predict", model_predict_s)
            imp_values = [
                self._random_state.choice(model.classes_, p=p, size=1)[0]
                for p in bachelor_preds
            ]
        else:
            # Collect the candidates, and the predictions for the candidates and bachelors
            model_predict_s = datetime.now()
            bachelor_preds = np.array(model.predict(bachelor_features))
            candidate_preds = np.array(model.predict(candidate_features))
            self.time_log.add_time("model_predict", model_predict_s)
            candidate_values = np.array(candidate_values)

            # Determine the nearest neighbors of the bachelor predictions in the candidate predictions
            knn = NearestNeighbors(n_neighbors=mmc, algorithm="ball_tree")
            knn.fit(candidate_preds.reshape(-1, 1))
            knn_indices = knn.kneighbors(
                bachelor_preds.reshape(-1, 1), return_distance=False
            )
            index_choice: List[int] = [
                self._random_state.choice(i) for i in knn_indices
            ]
            imp_values = candidate_values[index_choice]

        self.time_log.add_time("mean_match", mean_match_s)
        return imp_values

    # Models are updated here, and only here.
    def mice(self, iterations: int = 5, verbose: bool = False, **kw_fit):
        """
        Perform mice given dataset.

        Multiple Imputation by Chained Equations (MICE) is an
        iterative method which fills in (imputes) missing data
        points in a dataset by modeling each column using the
        other columns, and then inferring the missing data.

        For more information on MICE, and missing data in
        general, see Stef van Buuren's excellent online book:
        https://stefvanbuuren.name/fimd/ch-introduction.html

        For detailed usage information, see this project's
        README on the github repository:
        https://github.com/AnotherSamWilson/miceforest#The-MICE-Algorithm

        Parameters
        ----------
        iterations: int
            The number of iterations to run.
        verbose: bool
            Should information about the process
            be printed?
        kw_fit:
            Additional arguments to pass to
            sklearn.RandomForestRegressor and
            sklearn.RandomForestClassifier

        """

        mice_s = datetime.now()

        iterations_at_start = self.iteration_count()
        iter_range = range(
            iterations_at_start + 1, iterations_at_start + iterations + 1
        )

        # Required shut mypy up.
        assert isinstance(self.mean_match_candidates, Dict)

        for iteration in iter_range:
            if verbose:
                print(str(iteration) + " ", end="")
            for var in self.response_vars:
                if verbose:
                    print(" | " + var, end="")

                x, y = self._make_xy(var=var)
                non_missing_ind = self.na_where[var] == False
                candidate_features = x[non_missing_ind]
                candidate_values = y[non_missing_ind]

                fit_s = datetime.now()
                if var in self.categorical_variables:
                    current_model = _default_rf_classifier(
                        random_state=self._random_state, **kw_fit
                    )
                else:
                    current_model = _default_rf_regressor(
                        random_state=self._random_state, **kw_fit
                    )

                current_model.fit(X=candidate_features, y=candidate_values)
                self.time_log.add_time("model_fit", fit_s)

                self._insert_new_model(var=var, model=current_model)

                bachelor_features = x[self.na_where[var]]
                mmc = self.mean_match_candidates[var]

                if mmc == 0:
                    model_predict_s = datetime.now()
                    imp_values = current_model.predict(bachelor_features)
                    self.time_log.add_time("model_predict", model_predict_s)
                else:
                    is_categorical = var in self.categorical_variables
                    if is_categorical:
                        candidate_features = candidate_values = None
                    else:
                        ind = self.na_where[var] == False
                        candidate_features = x[ind]
                        candidate_values = y[ind]

                    imp_values = self._mean_match(
                        model=current_model,
                        mmc=mmc,
                        is_categorical=is_categorical,
                        bachelor_features=bachelor_features,
                        candidate_features=candidate_features,
                        candidate_values=candidate_values,
                    )

                self._insert_new_data(var, imp_values)
            if verbose:
                print("\n", end="")
        self.time_log.add_time("mice", mice_s)

    def impute_new_data(
        self,
        new_data: "DataFrame",
        iterations: int = None,
        save_all_iterations: bool = True,
        verbose: bool = False,
    ) -> ImputedDataSet:
        """
        Impute a new dataset

        Uses the models obtained while running MICE to impute new data,
        without fitting new models. Pulls mean matching candidates from
        the original data.

        save_models must be > 0. If save_models == 1, the last model
        obtained in mice is used for every iteration. If save_models > 1,
        the model obtained at each iteration is used to impute the new
        data for that iteration. If specified iterations is greater than
        the number of iterations run so far using mice, the last model
        is used for each additional iteration.

        Parameters
        ----------
        new_data: pandas DataFrame
            The new data to impute
        iterations: None, int
            The iterations to run. If None, the same number
            of iterations in the kernel are used. If iterations
            is greater than the number of MICE iterations, the
            latest model will be used for any additional iterations.
        save_all_iterations: bool
            Whether to save about all of the imputed values
            in every iteration, or just the last iteration.
        verbose: bool
            Print progress

        Returns
        -------
        ImputedDataSet

        """

        impute_new_data_s = datetime.now()

        if set(new_data.columns) != set(self.data.columns):
            raise ValueError("Columns are not the same as kernel data")

        if self.save_models < 1:
            raise ValueError("No models were saved.")

        imputed_data_set = ImputedDataSet(
            new_data,
            # copied because it can be edited if there are no
            # missing values in the new data response variables
            variable_schema=self.variable_schema.copy(),
            mean_match_candidates=self.mean_match_candidates,
            save_all_iterations=save_all_iterations,
            random_state=self._random_state,
        )

        curr_iters = self.iteration_count()
        iterations = self._default_iteration(iterations)
        iter_range = range(1, iterations + 1)
        iter_vars = imputed_data_set.response_vars
        assert isinstance(self.mean_match_candidates, Dict)

        for iteration in iter_range:
            if verbose:
                print("\n" + str(iteration) + " ", end="")

            # Determine which model iteration to grab
            if self.save_models == 1 or iteration > curr_iters:
                itergrab = curr_iters
            else:
                itergrab = iteration

            for var in iter_vars:
                if verbose:
                    if var == iter_vars[-1]:
                        endcap = "\n"
                    else:
                        endcap = ""
                    print(" | " + var, end=endcap)

                x, y = imputed_data_set._make_xy(var)
                kernelx, kernely = self._make_xy(var)
                bachelor_features = x[imputed_data_set.na_where[var]]
                mmc = self.mean_match_candidates[var]

                if mmc == 0:
                    imp_values = self.get_model(var, itergrab).predict(
                        bachelor_features
                    )
                else:
                    is_categorical = var in self.categorical_variables
                    if is_categorical:
                        candidate_features = candidate_values = None
                    else:
                        ind = self.na_where[var] == False
                        candidate_features = kernelx[ind]
                        candidate_values = kernely[ind]

                    imp_values = self._mean_match(
                        model=self.get_model(var, itergrab),
                        mmc=mmc,
                        is_categorical=is_categorical,
                        bachelor_features=bachelor_features,
                        candidate_features=candidate_features,
                        candidate_values=candidate_values,
                    )

                imputed_data_set._insert_new_data(var, imp_values)

        self.time_log.add_time("impute_new_data", impute_new_data_s)
        return imputed_data_set

    def get_feature_importance(self, iteration: int = None) -> DataFrame:
        """
        Return a matrix of feature importance. The cells
        represent the normalized feature importance of the
        columns to impute the rows. This is calculated
        internally by RandomForestRegressor/Classifier.

        Parameters
        ----------
        iteration: int
            The iteration to return the feature importance for.
            Right now, the model must be saved to return importance

        Returns
        -------
        pandas DataFrame
            A pandas DataFrame with variable column names and
            indexes.

        """
        # Should change this to save importance as models are updated, so
        # we can still get feature importance even if models are not saved.

        iteration = self._default_iteration(iteration)

        importance_matrix = DataFrame(
            columns=sorted(self.predictor_vars),
            index=sorted(self.response_vars),
            dtype=np.double,
        )

        for ivar in self.response_vars:
            importance_dict = dict(
                zip(
                    self.variable_schema[ivar],
                    self.get_model(ivar, iteration).feature_importances_,
                )
            )
            for pvar in importance_dict:
                importance_matrix.loc[ivar, pvar] = np.round(importance_dict[pvar], 3)

        return importance_matrix

    def plot_feature_importance(self, iteration: int = None, **kw_plot):
        """
        Plot the feature importance. See get_feature_importance()
        for more details.

        Parameters
        ----------
        iteration: int
            The iteration to plot the feature importance of.
        kw_plot
            Additional arguments sent to sns.heatmap()

        """
        import seaborn as sns

        importance_matrix = self.get_feature_importance(iteration=iteration)
        print(sns.heatmap(importance_matrix, **kw_plot))



from .ImputedDataSet import ImputedDataSet
from .ImputationSchema import _ImputationSchema
import numpy as np
from pandas import DataFrame
from itertools import combinations
from .utils import _distinct_from_list, _var_comparison
from typing import Union, List, Dict


class MultipleImputedDataSet(_ImputationSchema):
    """
    A collection of ImputedDataSets with similar schemas.
    Includes methods allowing for easy access and comparisons.
    Can be treated as a subscriptable list of ImputedDataSets.
    Stored datasets can be accessed by key in the range of:
    range(# Datasets).

    This class should not be instantiated by the user,
    but may be returned by a MultipleImputedKernel.impute_new_data()

    Parameters
    ----------
    initial_dataset: ImputedDataSet

    Methods
    -------
    append()
        Adds an imputed dataset.

    remove()
        Removes an imputed dataset.

    dataset_count()
        Returns the number of datasets being stored.

    get_correlations()
        Returns the pearson correlations between imputed datasets.

    plot_correlations()
        Plots the pearson correlations between imputed datasets.

    plot_imputed_distributions()
        Plots the distribution of imputed values against the original.

    iteration_count()
        Returns the iterations that have been run to impute the data.

    complete_data()
        Returns the completed data for the specified dataset

    plot_mean_convergence()
        Plots the means of the imputed values for each dataset over
        iterations.

    """

    def __init__(self, initial_dataset: ImputedDataSet):

        # Inherit schema and other attributes from initial_dataset
        super().__init__(
            validation_data=initial_dataset.data,
            variable_schema=initial_dataset.variable_schema,
            mean_match_candidates=initial_dataset.mean_match_candidates,
        )
        self.data = getattr(initial_dataset, "data")
        self.save_all_iterations = getattr(initial_dataset, "save_all_iterations")
        self.categorical_variables = getattr(initial_dataset, "categorical_variables")
        self._varfilter = getattr(initial_dataset, "_varfilter")
        self._prep_multi_plot = getattr(initial_dataset, "_prep_multi_plot")
        self._default_iteration = getattr(initial_dataset, "_default_iteration")

        self.imputed_data_sets = {0: initial_dataset}

    def __getitem__(self, key):
        return self.imputed_data_sets[key]

    def __setitem__(self, key, newitem):
        self.imputed_data_sets[key] = newitem

    def __delitem__(self, key):
        del self.imputed_data_sets[key]

    def items(self):
        return self.imputed_data_sets.items()

    def values(self):
        return self.imputed_data_sets.values()

    def keys(self):
        return self.imputed_data_sets.keys()

    def __repr__(self):
        summary_string = (
            f"""\
              Class: MultipleImputedDataSet\n"""
            + self._mids_info()
        )
        return summary_string

    def _mids_info(self) -> str:
        summary_string = f"""\
           Datasets: {self.dataset_count()}
         Iterations: {self.iteration_count()}
  Imputed Variables: {self.n_imputed_vars}
save_all_iterations: {self.save_all_iterations}"""
        return summary_string

    def _ensure_dataset_fidelity(self, new_set: ImputedDataSet):
        """
        To be consistent with the original, an imputed dataset must
        have the same:
            1) schema
            2) data
            3) number of iterations
            4) save_all_iterations

        Datasets can be updated internally, but cannot be
        added to the dict unless they are consistent.
        """

        assert self.equal_schemas(new_set)
        assert self.data.equals(new_set.data)
        assert new_set.iteration_count() == self.iteration_count()
        assert new_set.save_all_iterations == self.save_all_iterations

    def append(self, imputed_data_set: ImputedDataSet):
        """
        Appends an ImputedDataSet

        Parameters
        ----------
        imputed_data_set: ImputedDataSet
            The dataset to add

        """
        self._ensure_dataset_fidelity(imputed_data_set)
        curr_count = self.dataset_count()
        self[curr_count] = imputed_data_set

    def remove(self, datasets: Union[int, List[int]]):
        """
        Remove an ImputedDataSet by key. Renames keys
        in remaining datasets to be sequential.

        Parameters
        ----------
        datasets: int or list of int
            The dataset(s) to remove.

        """
        if isinstance(datasets, int):
            datasets = [datasets]

        for dataset in datasets:
            del self[dataset]

        # Rename the keys. Supporting 3.6 means we can't use zip,
        # or they'll be unordered.
        curr_keys = list(self.keys())
        for key in curr_keys:
            ind = curr_keys.index(key)
            self.imputed_data_sets[ind] = self.imputed_data_sets.pop(key)

    def dataset_count(self) -> int:
        """
        Returns the number of datasets being stored
        """
        return len(self.imputed_data_sets)

    def _get_all_vars(self) -> List[str]:
        all_vars = _distinct_from_list(
            [ids.all_vars for key, ids in self.items()], flatten=True
        )
        return all_vars

    def _get_cat_vars(self, response=True, predictor=False) -> List[str]:
        cat_vars = _distinct_from_list(
            [ids.categorical_variables for key, ids in self.items()], flatten=True
        )
        cat_vars = self._varfilter(cat_vars, response, predictor)
        return cat_vars

    def _get_num_vars(self, response=True, predictor=False) -> List[str]:
        all_vars = self._get_all_vars()
        cat_vars = self._get_cat_vars()
        num_vars = [i for i in all_vars if i not in cat_vars]
        num_vars = self._varfilter(num_vars, response, predictor)
        return num_vars

    def get_correlations(
        self, variables: List[str]
    ) -> Dict[str, Dict[int, List[float]]]:
        """
        Return the correlations between datasets for
        the specified variables.

        Parameters
        ----------
        variables: None, List[str]
            The variables to return the correlations for.

        Returns
        -------
        dict
            The correlations at each iteration for the specified
            variables.

        """

        if self.dataset_count() < 3:
            raise ValueError(
                "Not enough datasets to calculate correlations between them"
            )
        curr_iteration = self.iteration_count()

        # For every variable, get the correlations between every dataset combination
        # at each iteration
        correlation_dict = {}
        if self.save_all_iterations:
            iter_range = list(range(curr_iteration + 1))
        else:
            # Make this iterable for code tidyness
            iter_range = [curr_iteration]

        for var in variables:
            # Get a dict of variables and imputations for all datasets for this iteration
            iteration_level_imputations = {
                iteration: {
                    key: dataset[var, iteration] for key, dataset in self.items()
                }
                for iteration in iter_range
            }

            combination_correlations = {
                iteration: np.array(
                    [
                        round(np.corrcoef(impcomb)[0, 1], 3)
                        for impcomb in list(combinations(varimps.values(), 2))
                    ]
                )
                for iteration, varimps in iteration_level_imputations.items()
            }

            correlation_dict[var] = combination_correlations

        return correlation_dict

    # Helper methods that reach into self.imputed_data_sets
    def iteration_count(self, dataset: int = None, var: str = None):
        """
        Return iterations for the entire dataset, or a specific variable

        Parameters
        ----------
        dataset: int
            The dataset to return the iterations for. If None,
            the iteration for all datasets is returned. Will fail
            if iterations are not consistent between datasets.
        var: None,str
            If None, the iteration of all variables is returned.
            Will fail if iterations are not consistent.

        Returns
        -------
        int
            The iterations run so far.
        """
        if dataset is not None:
            self[dataset].iteration_count(var=var)
        else:
            # Get iterations for all imputed data sets.
            ids_iterations = _distinct_from_list(
                [ids.iteration_count(var=var) for key, ids in self.items()]
            )
            if len(ids_iterations) > 1:
                raise ValueError("Iterations are not consistent.")
            else:
                return next(iter(ids_iterations))

    def complete_data(
        self, dataset: int, iteration: int = None, all_vars: bool = False
    ) -> DataFrame:
        """
        Calls complete_data() from the specified stored dataset. See
        ImputedDataSet.complete_data().

        Parameters
        ----------
        dataset: int
            The dataset to return
        iteration:
            Iteration to return. If None, the latest iteration is
            returned. Iteration must have been saved if iteration
            is not None.
        all_vars: bool
            Should all variables used in the process be imputed,
            or just the ones specified as response variables?

        Returns
        -------
        pandas DataFrame
            The completed data
        """
        compdat = self[dataset].complete_data(iteration=iteration, all_vars=all_vars)
        return compdat

    def plot_mean_convergence(self, variables: List[str] = None, **adj_args):
        """
        Plots the average value of imputations over each iteration.

        Parameters
        ----------
        variables: List[str]
            The variables to plot. Must be numeric.
        adj_args
            Passed to matplotlib.pyplot.subplots_adjust()

        """
        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        if self.iteration_count() < 2:
            raise ValueError("There is only one iteration.")

        num_vars = self._get_num_vars()
        variables = _var_comparison(variables, num_vars)

        mean_dict = {key: ds.get_means(variables=variables) for key, ds in self.items()}
        plots, plotrows, plotcols = self._prep_multi_plot(variables)
        gs = gridspec.GridSpec(plotrows, plotcols)
        fig, ax = plt.subplots(plotrows, plotcols, squeeze=False)

        for v in range(plots):
            axr, axc = gs[v].get_rows_columns()[2], gs[v].get_rows_columns()[5]
            var = num_vars[v]
            for d in mean_dict.values():
                ax[axr, axc].plot(list(d[var].values()), color="black")
            ax[axr, axc].set_title(var)
            ax[axr, axc].set_xlabel("Iteration")
            ax[axr, axc].set_ylabel("mean")
        plt.subplots_adjust(**adj_args)

    def plot_imputed_distributions(
        self, variables: List[str] = None, iteration: int = None, **adj_args
    ):
        """
        Plot the imputed value distribution for all datasets.
        Red lines are the distribution of original data.
        Black lines are the distribution of the imputed values
        for each dataset.

        Parameters
        ----------
        variables: None,list
            The variables to plot.
        iteration: None,int
            The iteration to plot the distribution for.
            If None, the latest iteration is plotted.
        adj_args
            Additional arguments passed to plt.subplots_adjust()

        """
        import seaborn as sns
        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        iteration = self._default_iteration(iteration)
        num_vars = self._get_num_vars()
        variables = _var_comparison(variables, num_vars)

        plots, plotrows, plotcols = self._prep_multi_plot(variables)
        gs = gridspec.GridSpec(plotrows, plotcols)
        fig, ax = plt.subplots(plotrows, plotcols, squeeze=False)

        for v in range(plots):
            var = variables[v]
            gsgrc = gs[v].get_rows_columns()
            axr, axc = gsgrc[2], gsgrc[5]
            iteration_level_imputations = {
                key: dataset[var, iteration] for key, dataset in self.items()
            }
            plt.sca(ax[axr, axc])
            ax[axr, axc] = sns.distplot(
                self.data[var].dropna(),
                hist=False,
                kde=True,
                kde_kws={"linewidth": 2, "color": "red"},
            )
            for imparray in iteration_level_imputations.values():
                ax[axr, axc] = sns.distplot(
                    imparray,
                    hist=False,
                    kde=True,
                    kde_kws={"linewidth": 1, "color": "black"},
                )

        plt.subplots_adjust(**adj_args)

    def plot_correlations(self, variables: List[str] = None, **adj_args):
        """
        Plot the correlations between datasets.
        See get_correlations() for more details.

        Parameters
        ----------
        variables: None,list
            The variables to plot.
        adj_args
            Additional arguments passed to plt.subplots_adjust()

        """
        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        if self.dataset_count() < 4:
            raise ValueError("Not enough datasets to make box plot")

        num_vars = self._get_num_vars()
        variables = _var_comparison(variables, num_vars)
        plots, plotrows, plotcols = self._prep_multi_plot(variables)
        correlation_dict = self.get_correlations(variables=variables)
        gs = gridspec.GridSpec(plotrows, plotcols)
        fig, ax = plt.subplots(plotrows, plotcols, squeeze=False)

        for v in range(plots):
            axr, axc = gs[v].get_rows_columns()[2], gs[v].get_rows_columns()[5]
            var = list(correlation_dict)[v]
            ax[axr, axc].boxplot(
                list(correlation_dict[var].values()),
                labels=range(len(correlation_dict[var])),
            )
            ax[axr, axc].set_title(var)
            ax[axr, axc].set_xlabel("Iteration")
            ax[axr, axc].set_ylabel("Correlations")
            ax[axr, axc].set_ylim([-1, 1])
        plt.subplots_adjust(**adj_args)



import numpy as np
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from typing import List, Optional, Union, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pandas import DataFrame


def ampute_data(
    data: "DataFrame",
    variables: Optional[List[str]] = None,
    perc: float = 0.1,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
) -> "DataFrame":
    """
    Ampute Data

    Returns a copy of data with specified variables amputed.

    Parameters
    ----------
     data : Pandas DataFrame
        The data to ampute
     variables : None or list
        If None, are variables are amputed.
     perc : double
        The percentage of the data to ampute.
    random_state: None, int, or np.random.RandomState

    Returns
    -------
    pandas DataFrame
        The amputed data
    """
    amputed_data = data.copy()
    nrow = amputed_data.shape[0]
    amp_rows = int(perc * nrow)
    random_state = ensure_rng(random_state)

    if variables is None:
        variables = list(amputed_data.columns)

    for v in variables:
        na_ind = random_state.choice(range(nrow), replace=False, size=amp_rows)
        amputed_data.loc[na_ind, v] = np.NaN

    return amputed_data


def ensure_rng(
    random_state: Optional[Union[int, np.random.RandomState]] = None
) -> np.random.RandomState:
    """
    Creates a random number generator based on an optional seed.  This can be
    an integer or another random state for a seeded rng, or None for an
    unseeded rng.
    """
    if random_state is None:
        random_state = np.random.RandomState()
    elif isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)
    else:
        assert isinstance(random_state, np.random.RandomState)
    return random_state


# These exist so we can make a default classifier with the same parameters
# as those that may be passed to **kw_fit
def _default_rf_classifier(
    random_state: np.random.RandomState,
    max_features="sqrt",
    n_estimators=50,
    min_samples_leaf=1,
    bootstrap=True,
    max_samples=0.632,
    **kw_fit
) -> RandomForestClassifier:

    rfc = RandomForestClassifier(
        random_state=random_state,
        max_features=max_features,
        n_estimators=n_estimators,
        min_samples_leaf=min_samples_leaf,
        bootstrap=bootstrap,
        max_samples=max_samples,
        **kw_fit
    )
    return rfc


def _default_rf_regressor(
    random_state: np.random.RandomState,
    max_features="sqrt",
    n_estimators=50,
    min_samples_leaf=5,
    bootstrap=True,
    max_samples=0.632,
    **kw_fit
) -> RandomForestRegressor:

    rfc = RandomForestRegressor(
        random_state=random_state,
        max_features=max_features,
        n_estimators=n_estimators,
        min_samples_leaf=min_samples_leaf,
        bootstrap=bootstrap,
        max_samples=max_samples,
        **kw_fit
    )
    return rfc


def _var_comparison(variables: Optional[List[str]], comparison: List[str]) -> List[str]:
    """
    If variables is None, set it equal to the comparison list
    Else, make sure all of variables are in comparison list.
    """
    if variables is None:
        variables = comparison
    elif any([v not in comparison for v in variables]):
        raise ValueError("Action not permitted on supplied variables.")
    return variables


# Exists because list(set()) ruins reproducibility
def _distinct_from_list(
    lst: Union[np.ndarray, List[Any]], flatten: bool = False
) -> List[Any]:
    output = []

    # https://stackoverflow.com/questions/952914/how-to-make-a-flat-list-out-of-list-of-lists
    if flatten:
        lst = [item for sublist in lst for item in sublist]

    for item in lst:
        if item not in output:
            output.append(item)
    return output


def _copy_and_remove(lst, elements):
    lt = lst.copy()
    for element in elements:
        lt.remove(element)
    return lt


def _get_default_mmc(candidates=None):
    if candidates is None:
        return 5
    else:
        percent = 0.001
        minimum = 5
        mean_match_candidates = np.max(minimum, int(percent * candidates))
        return mean_match_candidates


def _list_union(a, b):
    return [element for element in a if element in b]


def _setequal(a, b):
    if not hasattr(a, "__iter__"):
        return a == b
    else:
        return set(a) == set(b)



from .ImputationSchema import _ImputationSchema
import numpy as np
from pandas import DataFrame
from .utils import (
    ensure_rng,
    _distinct_from_list,
    _copy_and_remove,
    _list_union,
    _var_comparison,
)
from typing import Optional, Union, List, Dict


class ImputedDataSet(_ImputationSchema):
    """
    Imputed Data Set

    This class should not be instantiated directly.
    Instead, use derived method MultipleImputedKernel.

    Parameters
    ----------
    data: DataFrame
        A pandas DataFrame to impute.

    variable_schema: None or list or dict
        If None all variables are used to impute all variables which have
        missing values.
        If list all variables are used to impute the variables in the list
        If dict the values will be used to impute the keys.

    mean_match_candidates:  None or int or dict
        The number of mean matching candidates to use. Mean matching
        allows the process to impute more realistic values.
        Candidates are _always_ drawn from a kernel dataset.
        Mean matching follows the following rules based on variable type:
            Categorical:
                If mmc = 0, the predicted class is used
                If mmc > 0, return class based on random draw weighted by
                    class probability for each sample.
            Numeric:
                If mmc = 0, the predicted value is used
                If mmc > 0, obtain the mmc closest candidate
                    predictions and collect the associated
                    real candidate values. Choose 1 randomly.

        For more information, see:
        https://github.com/AnotherSamWilson/miceforest#Predictive-Mean-Matching

    save_all_iterations: boolean, optional(default=False)
        Save all iterations that have been imputed, or just the latest.
        Saving all iterations allows for additional plotting,
        but may take more memory

    random_state: None,int, or numpy.random.RandomState
        Ensures a random state throughout the process

    Methods
    -------
    iteration_count()
        Returns the number of iterations that have been run to impute
        the data.

    complete_data()
        Returns a completed dataset, with missing values imputed.

    get_means()
        Returns the average value of numeric variables at each iteration.

    plot_mean_convergence()
        Plots the average value of numeric variables at each iteration.

    """

    def __init__(
        self,
        data: DataFrame,
        variable_schema: Union[List[str], Dict[str, List[str]]] = None,
        mean_match_candidates: Union[int, Dict[str, int]] = None,
        save_all_iterations: bool = True,
        random_state: Union[int, np.random.RandomState] = None,
    ):

        super().__init__(
            variable_schema=variable_schema,
            mean_match_candidates=mean_match_candidates,
            validation_data=data,
        )

        self._random_state = ensure_rng(random_state)

        self.data = data
        self.save_all_iterations = save_all_iterations
        self.categorical_variables = list(
            self.data_dtypes[self.data_dtypes == "category"].keys()
        )

        # Right now variables are filled in with random draws
        # from their original distribution. Add options in the future.
        self.imputation_values: Dict[str, Dict] = {
            var: dict() for var in self._all_imputed_vars
        }
        for var in self._all_imputed_vars:
            self.imputation_values[var] = {
                0: self._random_state.choice(
                    data[var].dropna(), size=self.na_counts[var]
                )
            }

    # Subsetting allows us to get to the imputation values:
    def __getitem__(self, tup):
        var, iteration = tup
        return self.imputation_values[var][iteration]

    def __setitem__(self, tup, newitem):
        var, iteration = tup
        self.imputation_values[var][iteration] = newitem

    def __delitem__(self, tup):
        var, iteration = tup
        del self.imputation_values[var][iteration]

    def __repr__(self):
        summary_string = " " * 14 + "Class: ImputedDataSet\n" + self._ids_info()
        return summary_string

    def _ids_info(self) -> str:
        summary_string = f"""\
         Iterations: {self.iteration_count()}
  Imputed Variables: {self.n_imputed_vars}
save_all_iterations: {self.save_all_iterations}"""
        return summary_string

    def iteration_count(self, var: str = None) -> int:
        """
        Return iterations for the entire dataset, or a specific variable

        Parameters
        ----------
        var: None,str
            If None, the meta iteration is returned.

        Returns
        -------
        int
            The iterations run so far.
        """

        # If var is None, we want the meta iteration level. This must fail
        # if called inside iteration updates, which would mean certain variables
        # have different numbers of iterations.
        if var is None:
            var_iterations = [
                np.max(list(itr))
                for var, itr in self.imputation_values.items()
                if var in self.response_vars
            ]
            distinct_iterations = _distinct_from_list(var_iterations)
            if len(distinct_iterations) > 1:
                raise ValueError(
                    "Inconsistent state - cannot get meta iteration count."
                )
            else:
                return next(iter(distinct_iterations))
        else:
            # Extract the number of iterations so far for a specific dataset, variable
            return np.max(list(self.imputation_values[var]))

    def _default_iteration(self, iteration: Optional[int], **kwargs) -> int:
        """
        If iteration is not specified it is assumed to
        be the last iteration run in many cases.
        """
        if iteration is None:
            return self.iteration_count(**kwargs)
        else:
            return iteration

    def _varfilter(self, vrs, response, predictor) -> List[str]:
        """
        Extracts predictor and response variables
        from a list of variables.
        """
        if not response and not predictor:
            return vrs
        if response:
            vrs = _list_union(vrs, self.response_vars)
        if predictor:
            vrs = _list_union(vrs, self.predictor_vars)
        return vrs

    def _get_cat_vars(self, response=True, predictor=False) -> List[str]:
        cat_vars = self._varfilter(
            vrs=self.categorical_variables, response=response, predictor=predictor
        )
        return cat_vars

    def _get_num_vars(self, response=True, predictor=False):
        num_vars = [v for v in self.data.columns if v not in self.categorical_variables]
        num_vars = self._varfilter(vrs=num_vars, response=response, predictor=predictor)
        return num_vars

    def _make_xy(self, var: str, iteration: int = None):
        """
        Make the predictor and response set used to train the model.
        Must be defined in ImputedDataSet because this method is called
        directly in KernelDataSet.impute_new_data()

        If iteration is None, it returns the most up-to-date imputations
        for each variable.
        """
        xvars = self.variable_schema[var]
        completed_data = self.complete_data(iteration=iteration, all_vars=True)
        to_convert = _list_union(self.categorical_variables, xvars)
        for ctc in to_convert:
            completed_data[ctc] = completed_data[ctc].cat.codes
        x = completed_data[xvars]
        y = completed_data[var]
        return x, y

    def _insert_new_data(self, var: str, new_data: np.ndarray):
        current_iter = self.iteration_count(var)
        if not self.save_all_iterations:
            del self[var, current_iter]
        self[var, current_iter + 1] = new_data

    def complete_data(self, iteration: int = None, all_vars: bool = False) -> DataFrame:
        """
        Replace missing values with imputed values.

        Parameters
        ----------
        iteration: int
            The iteration to return.
            If None, returns the most up-to-date iterations,
            even if different between variables.
            If not none, iteration must have been saved in
            imputed values.
        all_vars: bool
            Should all variables in the imputation schema be
            imputed, or just the ones specified to be imputed?

        Returns
        -------
        pandas DataFrame
            The completed data

        """
        imputed_dataframe = self.data.copy()

        # Need to impute all variables used in variable_schema if we are running model
        # Just impute specified variables if the user wants it.
        ret_vars = self._all_imputed_vars if all_vars else self.response_vars

        for var in ret_vars:
            itrn = self._default_iteration(iteration=iteration, var=var)
            imputed_dataframe.loc[self.na_where[var], var] = self[var, itrn]
        return imputed_dataframe

    def _cross_check_numeric(self, variables: Optional[List[str]]) -> List[str]:

        numeric_imputed_vars = _copy_and_remove(variables, self.categorical_variables)

        if variables is None:
            variables = numeric_imputed_vars
        else:
            if any([var not in numeric_imputed_vars for var in variables]):
                raise ValueError(
                    "Specified variable is not in imputed numeric variables."
                )

        return variables

    def get_means(self, variables: List[str] = None):
        """
        Return a dict containing the average imputation value
        for specified variables at each iteration.
        """
        num_vars = self._get_num_vars()
        variables = _var_comparison(variables, num_vars)

        # For every variable, get the correlations between every dataset combination
        # at each iteration
        curr_iteration = self.iteration_count()
        if self.save_all_iterations:
            iter_range = list(range(curr_iteration + 1))
        else:
            # Make this iterable for code tidyness
            iter_range = [curr_iteration]

        mean_dict = {
            var: {itr: np.mean(self[var, itr]) for itr in iter_range}
            for var in variables
        }
        return mean_dict

    def plot_mean_convergence(self, variables: List[str] = None, **adj_args):
        """
        Plots the average value of imputations over each iteration.

        Parameters
        ----------
        variables: List[str]
            The variables to plot. Must be numeric.
        adj_args
            Passed to matplotlib.pyplot.subplots_adjust()

        """
        if self.iteration_count() < 2 or not self.save_all_iterations:
            raise ValueError("There is only one iteration.")

        num_vars = self._get_num_vars()
        if variables is None:
            variables = num_vars
        elif any([v not in num_vars for v in variables]):
            raise ValueError("variables were either not numeric or not imputed.")

        mean_dict = self.get_means(variables=variables)

        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        plots = len(mean_dict)
        plotrows, plotcols = int(np.ceil(np.sqrt(plots))), int(
            np.ceil(plots / np.ceil(np.sqrt(plots)))
        )
        gs = gridspec.GridSpec(plotrows, plotcols)
        fig, ax = plt.subplots(plotrows, plotcols, squeeze=False)

        for v in range(plots):
            axr, axc = gs[v].get_rows_columns()[2], gs[v].get_rows_columns()[5]
            var = list(mean_dict)[v]
            ax[axr, axc].plot(list(mean_dict[var].values()))
            ax[axr, axc].set_title(var)
            ax[axr, axc].set_xlabel("Iteration")
            ax[axr, axc].set_ylabel("mean")
        plt.subplots_adjust(**adj_args)

    def _prep_multi_plot(
        self,
        variables: List[str],
    ):
        plots = len(variables)
        plotrows, plotcols = int(np.ceil(np.sqrt(plots))), int(
            np.ceil(plots / np.ceil(np.sqrt(plots)))
        )
        return plots, plotrows, plotcols

    def plot_imputed_distributions(
        self, variables: List[str] = None, iteration: int = None, **adj_args
    ):
        """
        Plot the imputed value distributions.
        Red lines are the distribution of original data
        Black lines are the distribution of the imputed values.

        Parameters
        ----------
        variables: None,list
            The variables to plot.
        iteration: None,int
            The iteration to plot the distribution for.
            If None, the latest iteration is plotted.
        adj_args
            Additional arguments passed to plt.subplots_adjust()

        """
        import seaborn as sns
        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        iteration = self._default_iteration(iteration)

        num_vars = self._get_num_vars()
        if variables is None:
            variables = num_vars
        elif any([v not in num_vars for v in variables]):
            raise ValueError("variables were either not numeric or not imputed.")

        plots, plotrows, plotcols = self._prep_multi_plot(variables)
        gs = gridspec.GridSpec(plotrows, plotcols)
        fig, ax = plt.subplots(plotrows, plotcols, squeeze=False)

        for v in range(plots):
            var = variables[v]
            gsgrc = gs[v].get_rows_columns()
            axr, axc = gsgrc[2], gsgrc[5]
            plt.sca(ax[axr, axc])
            ax[axr, axc] = sns.distplot(
                self.data[var].dropna(),
                hist=False,
                kde=True,
                kde_kws={"linewidth": 2, "color": "red"},
            )
            ax[axr, axc] = sns.distplot(
                self[var, iteration],
                hist=False,
                kde=True,
                kde_kws={"linewidth": 1, "color": "black"},
            )

        plt.subplots_adjust(**adj_args)



from numpy import concatenate
from .utils import _distinct_from_list, _copy_and_remove, _setequal, _list_union
from typing import Optional, Union, List, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from pandas import DataFrame


class _ImputationSchema:
    """
    Imputation Schema

    Contains information about how a dataset should be imputed.
    This class should not be instantiated directly.

    validation_data is the data that is to be imputed using this
    schema.

    variable_schema is validated and edited accordingly here

    mean_match_candidates are pulled from the kernel, so
    validation of mean_match_candidates is left to the kernel
    class, but stored here.
    """

    def __init__(
        self,
        validation_data: "DataFrame",
        variable_schema: Optional[Union[List[str], Dict[str, List[str]]]] = None,
        mean_match_candidates: Union[int, Dict[str, int]] = None,
    ):

        self.na_where = validation_data.isnull()
        self.na_counts = self.na_where.sum()
        self.data_dtypes = validation_data.dtypes
        self.data_shape = validation_data.shape
        self.data_variables = list(validation_data.columns)
        self.vars_with_any_missing = list(self.na_counts[self.na_counts > 0].keys())
        self.mean_match_candidates = mean_match_candidates

        if variable_schema is None:
            variable_schema = self.vars_with_any_missing

        if isinstance(variable_schema, list):
            variable_schema = {
                var: _copy_and_remove(self.data_variables, [var])
                for var in variable_schema
            }

        if isinstance(variable_schema, dict):
            self_impute_attempt = {
                key: (key in value) for key, value in variable_schema.items()
            }
            if any(self_impute_attempt.values()):
                self_impute_vars = [
                    key for key, value in self_impute_attempt.items() if value
                ]
                raise ValueError(
                    ",".join(self_impute_vars)
                    + " variables cannot be used to impute itself."
                )

        # Delete any variables that have no missing values.
        not_imputable = [
            var
            for var in list(variable_schema)
            if var not in self.vars_with_any_missing
        ]
        for rnm in not_imputable:
            del variable_schema[rnm]

        # Store values pertaining to variable schema
        self.variable_schema = variable_schema
        self.predictor_vars = _distinct_from_list(
            concatenate([value for key, value in variable_schema.items()])
        )
        self.response_vars = list(variable_schema)
        self.all_vars = _distinct_from_list(self.response_vars + self.predictor_vars)
        self._all_imputed_vars = _list_union(self.vars_with_any_missing, self.all_vars)
        self.n_imputed_vars = len(variable_schema)
        self.not_imputable = not_imputable

        # Get a list of variables that are being used to impute
        # other variables, but are not being imputed themselves.
        self.static_predictors = list(
            set(self.predictor_vars)
            & (set(self.vars_with_any_missing) - set(self.response_vars))
        )
        self.static_predictors.sort()

    def equal_schemas(self, imp_sch, fail=True):
        """
        Checks if two imputation schemas are similar enough
        """
        checks = {
            "response_vars": _setequal(self.response_vars, imp_sch.response_vars),
            "predictor_varsset": _setequal(self.predictor_vars, imp_sch.predictor_vars),
            "na_where": all(self.na_where == imp_sch.na_where),
            "mean_match_candidates": (
                _setequal(self.mean_match_candidates, imp_sch.mean_match_candidates)
            ),
        }
        failed_checks = [key for key, value in checks.items() if value is False]
        if len(failed_checks) > 0:
            if fail:
                raise ValueError(
                    "Inconsistency in schemas in regards to " + ",".join(failed_checks)
                )
            else:
                return False
        else:
            return True
